@@ -1,7 +1,7 @@
 /// \file
 // Range v3 library
 //
-//  Copyright Eric Niebler 2014
+//  Copyright Eric Niebler 2014-present
 //
 //  Use, modification and distribution is subject to the
 //  Boost Software License, Version 1.0. (See accompanying
@@ -89,18 +89,6 @@ namespace ranges
         {
             using reference_wrapper<T>::reference_wrapper;
         };
-        template<typename T, std::size_t Index>
-        struct indexed_element<T &, Index>
-          : reference_wrapper<T>
-        {
-            using reference_wrapper<T>::reference_wrapper;
-        };
-        template<typename T, std::size_t Index>
-        struct indexed_element<T &&, Index>
-          : reference_wrapper<T, true>
-        {
-            using reference_wrapper<T, true>::reference_wrapper;
-        };
         template<std::size_t Index>
         struct indexed_element<void, Index>
         {
@@ -175,10 +163,10 @@ namespace ranges
 
             template<typename T, typename Index>
             struct indexed_datum<T &, Index>
-              : reference_wrapper<T>
+              : reference_wrapper<T &>
             {
                 indexed_datum() = delete;
-                using reference_wrapper<T>::reference_wrapper;
+                using reference_wrapper<T &>::reference_wrapper;
                 constexpr indexed_element<T &, Index::value> ref() const noexcept
                 {
                     return {this->get()};
@@ -186,10 +174,10 @@ namespace ranges
             };
             template<typename T, typename Index>
             struct indexed_datum<T &&, Index>
-              : reference_wrapper<T, true>
+              : reference_wrapper<T &&>
             {
                 indexed_datum() = delete;
-                using reference_wrapper<T, true>::reference_wrapper;
+                using reference_wrapper<T &&>::reference_wrapper;
                 constexpr indexed_element<T &&, Index::value> ref() const noexcept
                 {
                     return {this->get()};
@@ -318,7 +306,7 @@ namespace ranges
             template<typename Fun, typename Proj = indexed_element_fn>
             constexpr int variant_visit_(std::size_t, variant_nil, Fun, Proj = {})
             {
-                return 0;
+                return (RANGES_EXPECT(false), 0);
             }
             template<typename Data, typename Fun, typename Proj = indexed_element_fn>
             constexpr int variant_visit_(std::size_t n, Data &self, Fun fun, Proj proj = {})
